#define	MINIMAL_SIZE
#ifdef MINIMAL_SIZE
#define	NO_RESOLVE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include <unistd.h>
#include <netdb.h>

#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/fcntl.h>

#define VERSION "2.0.0"

struct iopair
  {
    int		in, out;
    const char	*kill;
  };

static void
oops(const char *s)
{
  write(2, s, strlen(s));
  write(2, "\n", 1);
}

static void
ex(const char *s)
{
#ifdef MINIMAL_SIZE
  oops(s);
#else
  perror(s);
#endif
  exit(2);
}

union sockaddr_gen
   {
     struct sockaddr	sa;
     struct sockaddr_un	un;
     struct sockaddr_in	in;
  };

static int
getaddr(union sockaddr_gen *sin, const char *adr)
{
  char		*s, *host;
  int		max;
  size_t	len;

  len	= strlen(adr)+1;
  host	= alloca(len);
  memcpy(host, adr, len);

  memset(sin, 0, sizeof *sin);

  for (s=host; *s; s++)
    if (*s==':')
      {
	*s	= 0;

	sin->in.sin_family	= AF_INET;
	sin->in.sin_addr.s_addr	= htonl(INADDR_ANY);
	sin->in.sin_port	= htons(atoi(s+1));

	if (s!=host && !inet_aton(host, &sin->in.sin_addr))
	  {
#ifdef	NO_RESOLVE
	    ex(host);
#else
	    struct hostent	*he;

	    if ((he=gethostbyname(host))==0)
	      ex(host);
	    if (he->h_addrtype!=AF_INET || he->h_length!=sizeof sin->in.sin_addr)
	      ex("unsupported host address");
	    memcpy(&sin->in.sin_addr, he->h_addr, sizeof sin->in.sin_addr);
#endif
	  }
	return sizeof *sin;
      }

  sin->un.sun_family	= AF_UNIX;

  max = strlen(host);
  if (max > sizeof(sin->un.sun_path)-1)
    max = sizeof(sin->un.sun_path)-1;
  strncpy(sin->un.sun_path, host, max);

  return max + sizeof sin->un.sun_family;
}

static void
sockopen(struct iopair *io, const char *s, int do_connect)
{
  union sockaddr_gen	sin;
  int			on, len;
  struct linger		linger;
  int			sock;

  io->kill	= 0;
  len		= getaddr(&sin, s);

  sock	= socket(sin.sa.sa_family, SOCK_STREAM, 0);
  if (sock<0)
    ex("socket");

  on = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof on))
    ex("setsockopt reuse");

  linger.l_onoff	= 1;
  linger.l_linger	= 65535;
  if (setsockopt(sock, SOL_SOCKET, SO_LINGER, &linger, sizeof linger))
    ex("setsockopt linger");

  on	= 102400;
  if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &on, sizeof on))
    ex("setsockopt sndbuf");

  on	= 102400;
  if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &on, sizeof on))
    ex("setsockopt rcvbuf");

  if (!do_connect)
    {
      if (bind(sock, &sin.sa, len))
	ex("bind");
      if (sin.sa.sa_family==AF_UNIX)
	io->kill	= s;
    }

#if 0
  if (fcntl(sock, F_SETFL, O_NDELAY))
    ex("O_NDELAY");
#endif

  if (do_connect && connect(sock, &sin.sa, len))
    ex("connect");
  if (!do_connect && listen(sock, 1))
    ex("listen");

  io->in	= sock;
  io->out	= sock;
}

static void
docopy(int fi, int fo)
{
  char	buf[102400];
  int	got;

  while ((got=read(fi, buf, sizeof buf))!=0)
    {
      int	pos, put;

      if (got<0)
	{
	  if (errno==EINTR || errno==EAGAIN)
	    continue;
	  ex("input");
	}
      for (pos=0; pos<got; pos+=put)
	if ((put=write(fo, buf+pos, got-pos))<=0)
	  {
	    if (!put)
	      {
		oops("premature EOF on output");
		exit(3);
	      }
	    put	= 0;
	    if (errno==EINTR || errno==EAGAIN)
	      continue;
	    ex("output");
	  }
    }
}

static void
copyloop(struct iopair *one, struct iopair *two)
{
  pid_t	pid, a, b;
  int	st;

  if ((a=fork())==0)
    {
      docopy(one->in, two->out);
      exit(0);
    }
  if ((b=fork())==0)
    {
      docopy(two->in, one->out);
      exit(0);
    }
  st	= 4;
  while ((pid=wait(&st))==(pid_t)-1)
    {
      if (errno==ECHILD)
	break;
      ex("wait");
    }
  if (a!=pid && a!=(pid_t)-1)
    kill(a, SIGTERM);
  if (b!=pid && a!=(pid_t)-1)
    kill(b, SIGTERM);

  if (one->kill)
    unlink(one->kill);
  if (two->kill)
    unlink(two->kill);
  exit(st);
}


static void
getiopair(struct iopair *io, const char *s)
{
  union sockaddr_gen	sin;
  int			len, fd;

  switch (*s++)
    {
      default:
	ex("arg must start with < (accept) or > (connect)");

      case '<':
        sockopen(io, s, 0);

        len	= sizeof sin;
        if ((fd=accept(io->in, &sin.sa, &len))<0)
          ex("accept");

	close(io->in);
	io->in	= fd;
	io->out	= fd;
        break;

      case '>':
        sockopen(io, s, 1);
        break;
   }
}

int
main(int argc, char **argv)
{
  struct iopair one, two;

  two.in	= 0;
  two.out	= 1;
  two.kill	= 0;
  switch (argc)
    {
    case 3:
      getiopair(&two, argv[2]);
    case 2:
      getiopair(&one, argv[1]);
      break;

    default:
      oops("Usage: accept [SECOND] FIRST\n"
	    "Args:\t<SOCKET or >SOCKET\n"
	    "Socket:\t[host]:port or file");
      return 1;
    }
  copyloop(&one, &two);
  return 0;
}
